#include <assert.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <algorithm>
#include <map>
#include <queue>
#include <random>
#include <vector>
#include "Anthony.h"
#include "Catherine.h"

using std::vector;

void WrongAnswer(int id) {
  fprintf(stderr, "translate:wrong\n");
  fprintf(stderr, "Wrong Answer [%d]\n", id);
  puts("0.0");
  exit(0);
}

int N, M, A, B, S;
vector<int> U, V;

vector<vector<vector<int>>> adjacent_towns;
vector<int> distances;
std::mt19937 rng;
vector<vector<std::map<int, int>>> next_town_cache;

void InitManager(const vector<int> &x) {
  // adjacent_towns
  adjacent_towns.assign(N, vector<vector<int>>(A));
  for (int i = 0; i < M; ++i) {
    adjacent_towns[U[i]][x[i]].push_back(V[i]);
    adjacent_towns[V[i]][x[i]].push_back(U[i]);
  }
  for (int u = 0; u < N; ++u) {
    for (int j = 0; j < A; ++j) {
      std::shuffle(adjacent_towns[u][j].begin(), adjacent_towns[u][j].end(),
                   rng);
    }
  }
  // distances
  std::queue<int> que;
  distances.assign(N, -1);
  distances[0] = 0;
  que.push(0);
  for (; !que.empty(); ) {
    const int u = que.front();
    que.pop();
    for (int j = 0; j < A; ++j) {
      for (const int v : adjacent_towns[u][j]) {
        if (distances[v] == -1) {
          distances[v] = distances[u] + 1;
          que.push(v);
        }
      }
    }
  }
  // rng
  long long seed = 1LL * N * S * S;
  for (int u = 0; u < N; ++u) {
    seed ^= distances[u];
  }
  rng.seed(seed);
  // next_town_cache
  next_town_cache.assign(N, vector<std::map<int, int>>(A));
}

vector<vector<int>> GetNextTowns(int last_town, int current_town) {
  vector<vector<int>> next_towns(A);
  for (int j = 0; j < A; ++j) {
    for (const int v : adjacent_towns[current_town][j]) {
      if (v != last_town) {
        next_towns[j].push_back(v);
      }
    }
  }
  return next_towns;
}

int SelectNextTown(int last_town, int current_town, int z,
                   const vector<int> &vs) {
  const int vsLen = vs.size();
  int next_town = -1;
  switch (N % 3) {
    case 0: {
      // pure random
      next_town = vs[std::uniform_int_distribution<int>(0, vsLen - 1)(rng)];
      break;
    }
    case 1: {
      // shallower first, then cached random
      if (last_town == -1) {
        for (const int v : vs) {
          if (next_town == -1 || distances[next_town] > distances[v]) {
            next_town = v;
          }
        }
      } else {
        auto it = next_town_cache[current_town][z].find(last_town);
        if (it != next_town_cache[current_town][z].end()) {
          next_town = it->second;
        } else {
          next_town = vs[std::uniform_int_distribution<int>(0, vsLen - 1)(rng)];
        }
      }
      break;
    }
    case 2: {
      // deeper
      auto it = next_town_cache[current_town][z].find(last_town);
      if (it != next_town_cache[current_town][z].end()) {
        next_town = it->second;
      } else {
        for (const int v : vs) {
          if (next_town == -1 || distances[next_town] < distances[v]) {
            next_town = v;
          }
        }
      }
      break;
    }
    default: assert(false);
  }
  assert(next_town != -1);
  next_town_cache[current_town][z][last_town] = next_town;
  return next_town;
}

int main(int argc, char **argv) {
  if (argc != 5) {
    fprintf(stderr,
            "Usage: ./manager <u0-to-m> <m-to-u0> <u1-to-m> <m-to-u1>\n");
    exit(1);
  }
  signal(SIGPIPE, SIG_IGN);
  FILE *u0_to_m = fopen(argv[1], "r");
  FILE *m_to_u0 = fopen(argv[2], "w");
  FILE *u1_to_m = fopen(argv[3], "r");
  FILE *m_to_u1 = fopen(argv[4], "w");
  if (!u0_to_m || !m_to_u0 || !u1_to_m || !m_to_u1) {
    fprintf(stderr, "Error: Failed to open the pipes.\n");
    exit(1);
  }

  if (scanf("%d%d%d%d%d", &N, &M, &A, &B, &S) != 5) {
    fprintf(stderr, "Error: Failed to read data from stdin.\n");
    exit(1);
  }
  U.resize(M);
  V.resize(M);
  for (int i = 0; i < M; ++i) {
    if (scanf("%d%d", &U[i], &V[i]) != 2) {
      fprintf(stderr, "Error: Failed to read data from stdin.\n");
      exit(1);
    }
  }

  // Anthony
  // Mark(N, M, A, B, U, V)
  fprintf(m_to_u0, "%d %d %d %d\n", N, M, A, B);
  for (int i = 0; i < M; ++i) {
    fprintf(m_to_u0, "%d %d\n", U[i], V[i]);
  }
  fflush(m_to_u0);
  int flag;
  if (fscanf(u0_to_m, "%d", &flag) != 1) {
    fprintf(m_to_u1, "1\n");
    fflush(m_to_u1);
    WrongAnswer(0);
  }
  if (flag) {
    fprintf(m_to_u1, "1\n");
    fflush(m_to_u1);
    WrongAnswer(1);
  }
  vector<int> x(M);
  for (int i = 0; i < M; ++i) {
    if (fscanf(u0_to_m, "%d", &x[i]) != 1) {
      fprintf(m_to_u1, "1\n");
      fflush(m_to_u1);
      WrongAnswer(0);
    }
  }
  for (int i = 0; i < M; ++i) {
    if (!(0 <= x[i] && x[i] <= A - 1)) {
      fprintf(m_to_u1, "1\n");
      fflush(m_to_u1);
      WrongAnswer(2);
    }
  }

  // Catherine
  fprintf(m_to_u1, "0\n");
  fflush(m_to_u1);
  // Init(A, B)
  fprintf(m_to_u1, "%d %d\n", A, B);
  fflush(m_to_u1);
  InitManager(x);
  int num_moves = 0;
  int last_town = -1, current_town = S;
  for (; current_town != 0; ) {
    if (num_moves >= distances[S] + B) {
      fprintf(m_to_u1, "1\n");
      fflush(m_to_u1);
      WrongAnswer(6);
    }
    const vector<vector<int>> next_towns = GetNextTowns(last_town, current_town);
    vector<int> y(A);
    for (int j = 0; j < A; ++j) {
      y[j] = next_towns[j].size();
    }
    // move(y)
    fprintf(m_to_u1, "0\n");
    for (int j = 0; j < A; ++j) {
      fprintf(m_to_u1, "%d\n", y[j]);
    }
    fflush(m_to_u1);
    int z;
    if (fscanf(u1_to_m, "%d", &z) != 1) {
      WrongAnswer(0);
    }
    ++num_moves;
    if (!(-1 <= z && z <= A - 1)) {
      fprintf(m_to_u1, "1\n");
      fflush(m_to_u1);
      WrongAnswer(3);
    }
    if (last_town == -1 && z == -1) {
      fprintf(m_to_u1, "1\n");
      fflush(m_to_u1);
      WrongAnswer(4);
    }
    if (0 <= z && z <= A - 1 && y[z] == 0) {
      fprintf(m_to_u1, "1\n");
      fflush(m_to_u1);
      WrongAnswer(5);
    }
    if (z == -1) {
      std::swap(last_town, current_town);
    } else {
      const int next_town =
          SelectNextTown(last_town, current_town, z, next_towns[z]);
      last_town = current_town;
      current_town = next_town;
    }
  }
  fprintf(m_to_u1, "1\n");
  fflush(m_to_u1);
  fprintf(stderr, "translate:success\n");
  fprintf(stderr, "Accepted %d %d\n", distances[S], num_moves);
  puts("1.0");
  return 0;
}
